@@ -62,7 +62,19 @@ import <- function(module, as = NULL, convert = TRUE, delay_load = FALSE) {
     ensure_python_initialized(required_module = module)
 
     # import the module
-    py_module_import(module, convert = convert)
+    hookName <- paste("reticulate", module, "load", sep = "::")
+    imported <- py_module_import(module, convert = convert)
+
+    # run load hooks
+    hooks <- getHook(hookName)
+    for (hook in hooks)
+      tryCatch(hook(imported), error = warning)
+
+    # remove hooks (we only want to run on first import)
+    setHook(hookName, list(), "replace")
+
+    # return imported module
+    imported
   }
 
   # delay load case (wait until first access)
@@ -155,6 +167,19 @@ as.character.python.builtin.object <- function(x, ...) {
   py_str(x)
 }
 
+#' Convert Python bytes to an R character vector
+#'
+#' @inheritParams base::as.character
+#'
+#' @param encoding Encoding to use for conversion (defaults to utf-8)
+#' @param errors Policy for handling conversion errors. Default is 'strict'
+#'  which raises an error. Other possible values are 'ignore' and 'replace'
+#'
+#' @export
+as.character.python.builtin.bytes <- function(x, encoding = "utf-8", errors = "strict", ...) {
+  x$decode(encoding = encoding, errors = errors)
+}
+
 #' @export
 "==.python.builtin.object" <- function(a, b) {
   py_compare(a, b, "==")
@@ -224,27 +249,36 @@ py_has_convert <- function(x) {
 }
 
 py_maybe_convert <- function(x, convert) {
-  if (convert || py_is_callable(x)) {
 
-    # capture previous convert for attr
-    attrib_convert <- py_has_convert(x)
+  # if this is already an R object, nothing to do
+  if (!inherits(x, "python.builtin.object"))
+    return(x)
 
-    # temporarily change convert so we can call py_to_r and get S3 dispatch
-    envir <- as.environment(x)
-    assign("convert", convert, envir = envir)
-    on.exit(assign("convert", attrib_convert, envir = envir), add = TRUE)
+  # if it's neither convertable nor callable,
+  # nothing to do
+  convertable <- convert || py_is_callable(x)
+  if (!convertable)
+    return(x)
 
-    # call py_to_r
-    x <- py_to_r(x)
-  }
+  # perform conversion
+  # capture previous convert for attr
+  attrib_convert <- py_has_convert(x)
+
+  # temporarily change convert so we can call py_to_r and get S3 dispatch
+  envir <- as.environment(x)
+  assign("convert", convert, envir = envir)
+  on.exit(assign("convert", attrib_convert, envir = envir), add = TRUE)
+
+  # call py_to_r
+  py_to_r(x)
 
-  x
 }
 
 # helper function for accessing attributes or items from a
 # Python object, after validating that we do indeed have
 # a valid Python object reference
 py_get_attr_or_item <- function(x, name, prefer_attr) {
+
   # resolve module proxies
   if (py_is_module_proxy(x))
     py_resolve_module_proxy(x)
@@ -273,16 +307,27 @@ py_get_attr_or_item <- function(x, name, prefer_attr) {
   }
 
   # get the attrib and convert as needed
+  object <- NULL
   if (prefer_attr) {
     object <- py_get_attr(x, name)
   } else {
 
     # if we have an attribute, attempt to get the item
-    # but allow for fallback to that attribute
+    # but allow for fallback to that attribute. note that
+    # the logic here is fairly convoluted but is necessary
+    # to maintain backwards compatibility with a number of
+    # CRAN packages (hopefully we can simplify this in the
+    # future)
     if (py_has_attr(x, name)) {
-      object <- py_get_item(x, name, silent = TRUE)
+
+      # try to get item
+      if (py_has_attr(x, "__getitem__"))
+        object <- py_get_item(x, name, silent = TRUE)
+
+      # fallback to attribute
       if (is.null(object))
         object <- py_get_attr(x, name)
+
     } else {
       # we don't have an attribute; only attempt item
       # access and allow normal error propagation
@@ -326,10 +371,6 @@ as.environment.python.builtin.object <- function(x) {
 }
 
 
-#' @export
-`[[.python.builtin.object` <- `$.python.builtin.object`
-
-
 #' @export
 `$<-.python.builtin.object` <- function(x, name, value) {
   if (!py_is_null_xptr(x) && py_available())
@@ -399,9 +440,15 @@ as.environment.python.builtin.object <- function(x) {
     }
   }
 
+  idx <- grepl(pattern, names)
+  names <- names[idx]
+  types <- types[idx]
+
   if (length(names) > 0) {
     # set types
-    attr(names, "types") <- types
+    oidx <- order(names)
+    names <- names[oidx]
+    attr(names, "types") <- types[oidx]
 
     # specify a help_handler
     attr(names, "helpHandler") <- "reticulate:::help_handler"
@@ -466,7 +513,7 @@ plot.numpy.ndarray <- function(x, y, ...) {
 #'
 #' @return A Python dictionary
 #'
-#' @note The returned dictionary will not automatically convert it's elements
+#' @note The returned dictionary will not automatically convert its elements
 #'   from Python to R. You can do manual converstion with the [py_to_r()]
 #'   function or pass `convert = TRUE` to request automatic conversion.
 #'
@@ -529,7 +576,7 @@ py_dict <- function(keys, values, convert = FALSE) {
 #' @param ... Values for tuple (or a single list to be converted to a tuple).
 #'
 #' @return A Python tuple
-#' @note The returned tuple will not automatically convert it's elements from
+#' @note The returned tuple will not automatically convert its elements from
 #'   Python to R. You can do manual converstion with the [py_to_r()] function or
 #'   pass `convert = TRUE` to request automatic conversion.
 #'
@@ -735,13 +782,15 @@ iterate <- function(it, f = base::identity, simplify = TRUE) {
 #' @rdname iterate
 #' @export
 iter_next <- function(it, completed = NULL) {
-
-  # check for iterator
-  if (!inherits(it, "python.builtin.iterator"))
-    stop("iterator function called with non-iterator argument", call. = FALSE)
-
-  # call iterator
+  
+  # TODO: would like to use PyIter_Check() but that is only implemented
+  # as a macro in Python 2.x and requires copying more headers
+  iterable <- py_has_attr(it, "__next__") || py_has_attr(it, "next")
+  if (!iterable)
+    stop("object is not iterable", call. = FALSE)
+  
   py_iter_next(it, completed)
+  
 }
 
 
@@ -858,17 +907,14 @@ py_get_item <- function(x, key, silent = FALSE) {
   ensure_python_initialized()
   if (py_is_module_proxy(x))
     py_resolve_module_proxy(x)
-
-  if (!py_has_attr(x, "__getitem__"))
-    stop("Python object has no '__getitem__' method", call. = FALSE)
-  getitem <- py_to_r(py_get_attr(x, "__getitem__", silent = FALSE))
-
-  item <- if (silent)
-    tryCatch(getitem(key), error = function(e) NULL)
-  else
-    getitem(key)
-
-  item
+ 
+  # NOTE: for backwards compatibility, we make sure to return an R NULL on error 
+  if (silent) {
+    tryCatch(py_get_item_impl(x, key, FALSE), error = function(e) NULL)
+  } else {
+    py_get_item_impl(x, key, FALSE)
+  }
+  
 }
 
 #' Set an item for a Python object
@@ -890,12 +936,7 @@ py_set_item <- function(x, name, value) {
   ensure_python_initialized()
   if (py_is_module_proxy(x))
     py_resolve_module_proxy(x)
-
-  if (!py_has_attr(x, "__setitem__"))
-    stop("Python object has no '__setitem__' method", call. = FALSE)
-  setitem <- py_to_r(py_get_attr(x, "__setitem__", silent = FALSE))
-
-  setitem(name, value)
+  py_set_item_impl(x, name, value)
   invisible(x)
 }
 
@@ -1099,49 +1140,30 @@ py_capture_output <- function(expr, type = c("stdout", "stderr")) {
 
   # get output tools helper functions
   output_tools <- import("rpytools.output")
-
-  # handle stdout
-  restore_stdout <- NULL
-  if ("stdout" %in% type) {
-    restore_stdout <- output_tools$start_stdout_capture()
-    on.exit({
-      if (!is.null(restore_stdout))
-        output_tools$end_stdout_capture(restore_stdout)
-    }, add = TRUE)
-  }
-
-  # handle stderr
-  restore_stderr <- NULL
-  if ("stderr" %in% type) {
-    restore_stderr <- output_tools$start_stderr_capture()
-    on.exit({
-      if (!is.null(restore_stderr))
-        output_tools$end_stderr_capture(restore_stderr)
-    }, add = TRUE)
-  }
+  
+  # scope output capture
+  capture_stdout <- "stdout" %in% type
+  capture_stderr <- "stderr" %in% type
+  output_tools$start_capture(capture_stdout, capture_stderr)
+  on.exit(output_tools$end_capture(capture_stdout, capture_stderr), add = TRUE)
 
   # evaluate the expression
   force(expr)
 
-  # collect the output
-  output <- ""
-  if (!is.null(restore_stdout)) {
-    std_out <- output_tools$end_stdout_capture(restore_stdout)
-    output <- paste0(output, std_out)
-    if (nzchar(std_out))
-      output <- paste0(output, "\n")
-    restore_stdout <- NULL
-  }
-  if (!is.null(restore_stderr)) {
-    std_err <- output_tools$end_stderr_capture(restore_stderr)
-    output <- paste0(output, std_err)
-    if (nzchar(std_err))
-      output <- paste0(output, "\n")
-    restore_stderr <- NULL
-  }
+  # collect output
+  output_tools$collect_output()
+  
+}
+
+py_flush_output <- function() {
+
+  if (!is_python3())
+    return()
+
+  sys <- import("sys", convert = TRUE)
+  sys$stdout$flush()
+  sys$stderr$flush()
 
-  # return the output
-  output
 }
 
 
@@ -1166,6 +1188,7 @@ py_capture_output <- function(expr, type = c("stdout", "stderr")) {
 #' @export
 py_run_string <- function(code, local = FALSE, convert = TRUE) {
   ensure_python_initialized()
+  on.exit(py_flush_output(), add = TRUE)
   invisible(py_run_string_impl(code, local, convert))
 }
 
@@ -1173,6 +1196,7 @@ py_run_string <- function(code, local = FALSE, convert = TRUE) {
 #' @export
 py_run_file <- function(file, local = FALSE, convert = TRUE) {
   ensure_python_initialized()
+  on.exit(py_flush_output(), add = TRUE)
   invisible(py_run_file_impl(file, local, convert))
 }
 
@@ -1184,47 +1208,46 @@ py_eval <- function(code, convert = TRUE) {
 }
 
 py_callable_as_function <- function(callable, convert) {
+  
+  force(callable)
+  force(convert)
+  
   function(...) {
+    
     dots <- py_resolve_dots(list(...))
     result <- py_call_impl(callable, dots$args, dots$keywords)
-    if (convert) {
+    
+    if (convert)
       result <- py_to_r(result)
-      if (is.null(result))
-        invisible(result)
-      else
-        result
-    }
-    else {
+      
+    if (is.null(result))
+      invisible(result)
+    else
       result
-    }
+    
   }
+  
+}
+
+py_resolve_formals <- function(callback) {
+  
+  object <- attr(callback, "py_object")
+  if (!inherits(object, "python.builtin.object"))
+    return(NULL)
+  
+  tryCatch(py_get_formals(object), error = function(e) NULL)
+  
 }
 
 py_resolve_dots <- function(dots) {
-  args <- list()
-  keywords <- list()
-  names <- names(dots)
-  if (!is.null(names)) {
-    for (i in 1:length(dots)) {
-      name <- names[[i]]
-      if (nzchar(name))
-        if (is.null(dots[[i]]))
-          keywords[name] <- list(NULL)
-        else
-          keywords[[name]] <- dots[[i]]
-        else
-          if (is.null(dots[[i]]))
-            args[length(args) + 1] <- list(NULL)
-          else
-            args[[length(args) + 1]] <- dots[[i]]
-    }
-  } else {
-    args <- dots
-  }
-  list(
-    args = args,
-    keywords = keywords
-  )
+  
+  nms <- names(dots)
+  if (is.null(nms))
+    return(list(args = dots, keywords = list()))
+  
+  named <- nzchar(nms)
+  list(args = dots[!named], keywords = dots[named])
+  
 }
 
 
@@ -1332,7 +1355,8 @@ py_inject_r <- function(envir) {
 
   # define the getters, setters we'll attach to the Python class
   getter <- function(self, code) {
-    r_to_py(eval(parse(text = as_r_value(code)), envir = envir))
+    object <- eval(parse(text = as_r_value(code)), envir = envir)
+    r_to_py(object, convert = is.function(object))
   }
 
   setter <- function(self, name, value) {
@@ -1348,3 +1372,29 @@ py_inject_r <- function(envir) {
   py_run_string("r = R()")
 
 }
+
+py_inject_hooks <- function() {
+  
+  builtins <- import_builtins(convert = TRUE)
+  
+  input <- function(prompt) {
+    
+    response <- tryCatch(
+      readline(prompt),
+      interrupt = identity
+    )
+    
+    if (inherits(response, "interrupt"))
+      stop("KeyboardInterrupt", call. = FALSE)
+    
+    r_to_py(response)
+    
+  }
+  
+  # override input function
+  if (interactive()) {
+    name <- if (is_python3()) "input" else "raw_input"
+    builtins[[name]] <- input
+  }
+  
+}
